#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <asm/types.h>
#include <netdb.h>
#include "cmdsrv.h"

int main (int argc, char *argv[])
{
	int retcode;
    int sockfd;
    int len;
    struct sockaddr_in address;
    int result;
    CommandPacketT command;
	int reqLen;
    ResponsePacketT response;
	int byteswritten;
	int bytesread;
	int doGetMore;
	int GetMoreOffset;
	char *hostName;
	struct hostent *hostInfo;

	// Assume failure
	retcode = -1;
	sockfd = -1;
	hostInfo = NULL;

	// Did the user specify where the server is?
	if (argc != 1)
		hostName = argv[1];
	else
		hostName = "localhost";

	// Get the server's host info structure
	hostInfo = gethostbyname (hostName);
	if (hostInfo == NULL)
		goto HostError;

	// Create a client socket and connect to the server
    sockfd = socket (AF_INET, SOCK_STREAM, 0);
	if (sockfd == -1)
		goto SocketError;
    address.sin_family = AF_INET;
//    address.sin_addr.s_addr = inet_addr ("127.0.0.1");
    address.sin_addr = *(struct in_addr *)*hostInfo->h_addr_list;
    address.sin_port = htons (9734);
    len = sizeof(address);
    result = connect (sockfd, (struct sockaddr *)&address, len);
    if (result == -1)
		goto ConnectError;

	// Process commands from the user
	doGetMore = 0;
	GetMoreOffset = 0;
	do
	{
		// Init command packet
		memset (&command, 0, sizeof(CommandPacketT));

		// Do we need to get more data from the last command?
		if (doGetMore == 0)
		{
			// Get a command from the user
			printf ("\bPlease type a command line (ENTER to terminate) > ");
			fgets (command.u.ExecuteCommandLine, MAX_COMMAND_LINE-1, stdin);
			printf ("\n");

			// Determine if we need to stop
			reqLen = strlen(command.u.ExecuteCommandLine);
			if (reqLen == 1)
				reqLen = 0;
			if (reqLen == 0)
				continue;
			command.u.ExecuteCommandLine[reqLen-1] = '\0';	// Strip '\n'

			// Build the execute command packet
			command.Command = CMDSRV_CMD_EXECUTE;
		}
		else
		{
			// Build the get more command packet
			command.Command = CMDSRV_CMD_GET_MORE;
			command.u.GetMoreOffset = GetMoreOffset;
		}

		// Send a command request to the server
	    byteswritten = write (sockfd, &command, sizeof(CommandPacketT));
		if (byteswritten != sizeof(CommandPacketT))
			goto WriteError;

		// Read back the response
	    bytesread = read (sockfd, &response, sizeof(ResponsePacketT));
	    if (bytesread != sizeof(ResponsePacketT))
	    	goto ReadError;

		// Display response
	    printf ("CLIENT:  Response from server = %d, \"%s\", bytes read = %d\n", response.Response, response.ResponseBuffer, bytesread);

		// Do we need to do a get more command?
		if (response.Response == CMDSRV_RSP_OK_GET_MORE)
		{
			doGetMore = 1;
			GetMoreOffset += MAX_RESPONSE_BUFFER;
		}
		else
		{
			doGetMore = 0;
			GetMoreOffset = 0;
		}

	} while (reqLen != 0);

	// OK
	retcode = 0;

Exit:
	// Deallocate resources
	if (sockfd != -1)
	    close (sockfd);

	printf ("CLIENT:  Client exiting with return code = %d\n", retcode);
    return (retcode);

HostError:
	printf ("CLIENT:  ERROR:  Cannot resolve host address to %s!\n", hostName);
	retcode = 1;
	goto Exit;

SocketError:
	printf ("CLIENT:  ERROR:  Cannot open client socket!\n");
	retcode = 2;
	goto Exit;

ConnectError:
	printf ("CLIENT:  ERROR:  Cannot connect to server!\n");
	retcode = 3;
	goto Exit;

WriteError:
	printf ("CLIENT:  ERROR:  Cannot write to server!\n");
	retcode = 4;
	goto Exit;

ReadError:
	printf ("CLIENT:  ERROR:  Cannot read from server!\n");
	retcode = 5;
	goto Exit;

}
