#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include "cmdsrv.h"

void ServiceRequest (PCommandPacketT Command, PResponsePacketT Response);
int ExecuteCommandLine (char *CommandLine, char *OutputBuffer, int OutputBufferMaxSize, int *OutputBufferSize);
int GetMore (int GetMoreOffset, char *OutputBuffer, int OutputBufferMaxSize, int *OutputBufferSize);

int main()
{
	int retcode;
    int server_sockfd, client_sockfd;
    int server_len, client_len;
    struct sockaddr_in server_address;
    struct sockaddr_in client_address;
    int result;
    fd_set readfds;
    fd_set testfds;

	// Assume failure
	retcode = -1;

	// Create a named server socket
    server_sockfd = socket(AF_INET, SOCK_STREAM, 0);
    server_address.sin_family = AF_INET;
    server_address.sin_addr.s_addr = htonl(INADDR_ANY);
    server_address.sin_port = htons(9734);
    server_len = sizeof(server_address);
    bind (server_sockfd, (struct sockaddr *)&server_address, server_len);

	// Listen for clients on the server socket
	// MAX of 5 clients
    listen (server_sockfd, 5);

	// Initialize read fds to handle input from the server socket
    FD_ZERO (&readfds);
    FD_SET (server_sockfd, &readfds);

	// Main server loop
    while (1)
	{
        CommandPacketT command;
        ResponsePacketT response;
        int fd;
        int nread;

        printf("SERVER:  waiting\n");

		// Wait for a client to connect on the server socket
        testfds = readfds;
        result = select (
					FD_SETSIZE, 
					&testfds, 
					(fd_set *)0, 
					(fd_set *)0, 
					(struct timeval *) 0);
        if (result < 1) 
			goto SelectError;

		// Process all active descriptors on the server socket
        for (fd = 0; fd < FD_SETSIZE; fd++)
		{
			// Is the current descriptor active?
            if (FD_ISSET(fd,&testfds))
			{
				// Is it a connect request on the server socket?
                if(fd == server_sockfd)
				{
					// Accept the connection
                    client_sockfd = accept (
										server_sockfd, 
                        				(struct sockaddr *)&client_address, 
										&client_len);

					// Add client socket to our list of client descriptors
                    FD_SET (client_sockfd, &readfds);
                    printf ("SERVER:  adding client on fd %d\n", client_sockfd);
                }
                else
				{
					// Read the request from the client socket
                    ioctl(fd, FIONREAD, &nread);
					
					// Is it a "close" request?
                    if (nread == 0)
					{
						// Close the client socket
                        close (fd);

						// Remove the client socket from our list of client descriptors
                        FD_CLR(fd, &readfds);
                        printf("SERVER:  removing client on fd %d\n", fd);
                    }
                    else
					{
						int bytesread;
						int byteswritten;

						printf ("SERVER:  serving client on fd %d\n", fd);

						// Read the command "request"
                        bytesread = read (fd, &command, sizeof(CommandPacketT));
						if (bytesread == 0)
							continue;
						if (bytesread != sizeof(CommandPacketT))
							goto InvalidRequest;

						// Service the request
						ServiceRequest (&command, &response);

						// Send the response
                        byteswritten = write (fd, &response, sizeof(ResponsePacketT));
						if (byteswritten != sizeof(ResponsePacketT))
							printf ("SERVER:  Bytes written %d != %d\n", byteswritten, sizeof(ResponsePacketT));

                    } // else
                } // else
            } //if
        } //for
    } //while

	// OK
	retcode = 0;

Exit:

	// Clean up and return
	close (client_sockfd);
	close (server_sockfd);
	printf ("SERVER:  Server exiting with return code = %d\n", retcode);
	return (retcode);

SelectError:
	printf ("SERVER:  Select failed with status of %d\n", result);
	retcode = 1;
	goto Exit;

InvalidRequest:
	printf ("SERVER:  Invalid request received\n");
	retcode = 1;
	goto Exit;

}

void ServiceRequest (PCommandPacketT Command, PResponsePacketT Response)
{
	int ccode;

	printf ("SERVER:  command = %d\n", Command->Command);

	// Initialize response
	memset (Response, 0, sizeof(ResponsePacketT));
	Response->Response = CMDSRV_RSP_GENERAL_ERROR;

	// Service
	switch (Command->Command)
	{
		case CMDSRV_CMD_NOP:
			break;
		case CMDSRV_CMD_EXECUTE:
			ccode = ExecuteCommandLine (
						Command->u.ExecuteCommandLine, 
						Response->ResponseBuffer, 
						MAX_RESPONSE_BUFFER,
						&Response->ResponseBufferSize);
			if (ccode == 1)
				goto GetMore;
			if (ccode != 0)
				goto ExecuteError;
			break;
		case CMDSRV_CMD_GET_MORE:
			ccode = GetMore (
						Command->u.GetMoreOffset, 
						Response->ResponseBuffer, 
						MAX_RESPONSE_BUFFER,
						&Response->ResponseBufferSize);
			if (ccode == 1)
				goto GetMore;
			if (ccode != 0)
				goto GetMoreError;
			break;
		default:
			goto InvalidCommand;
			break;
	}

	// OK
	Response->Response = CMDSRV_RSP_OK;

Exit:
	printf ("SERVER:  response being returned:  %d\n", Response->Response);
	return;

InvalidCommand:
	printf ("SERVER:  Invalid command received, command = %d\n", Command->Command);
	Response->Response = CMDSRV_RSP_INVALID_COMMAND;
	goto Exit;

GetMore:
	Response->Response = CMDSRV_RSP_OK_GET_MORE;
	goto Exit;

ExecuteError:
	printf ("SERVER:  Could not execute the command line \"%s\"\n", Command->u.ExecuteCommandLine);
	Response->Response = CMDSRV_RSP_EXECUTE_ERROR;
	goto Exit;

GetMoreError:
	printf ("SERVER:  Could not get more data\n");
	Response->Response = CMDSRV_RSP_GET_MORE_ERROR;
	goto Exit;
}	

int ExecuteCommandLine (char *CommandLine, char *OutputBuffer, int OutputBufferMaxSize, int *OutputBufferSize)
{
	#define TEMP_FILE_NAME "__temp__.tmp"
	int retcode;
	char *cmdLineStr;
	int ccode;
	int fd;
	struct stat filestat;

	// Assume failure
	retcode = -1;
	cmdLineStr = NULL;
	fd = -1;
	*OutputBufferSize = 0;

	// Execute the command line and capture to a temp file
	cmdLineStr = malloc (strlen(CommandLine)+3+strlen(TEMP_FILE_NAME)+1);
	if (cmdLineStr == NULL)
		goto AllocError;
	strcpy (cmdLineStr, CommandLine);
	strcat (cmdLineStr, " > ");
	strcat (cmdLineStr, TEMP_FILE_NAME);
	ccode = system (cmdLineStr);
	if (ccode != 0)
		goto ExecError;

	// Determine the length of the temp file
	ccode = stat (TEMP_FILE_NAME, &filestat);
	if (ccode == -1)
		goto StatError;

	// Read the results
	fd = open (TEMP_FILE_NAME, O_RDONLY);
	if (fd == -1)
		goto OpenError;
	ccode = read (fd, OutputBuffer, OutputBufferMaxSize);
	if (ccode == -1)
		goto ReadError;
	*OutputBufferSize = ccode;

	// OK
	if (filestat.st_size > OutputBufferMaxSize)
		retcode = 1;	// Indicate that client needs to do GET_MORE commands
	else
		retcode = 0;

Exit:
	// Deallocate resources and return
	if (fd != -1)
		close (fd);
	if (cmdLineStr != NULL)
		free (cmdLineStr);
	return (retcode);

AllocError:
	printf ("ERROR:  Memory allocation error\n");
	retcode = 2;
	goto Exit;

ExecError:
	printf ("ERROR:  Error occurrred while executing command, code = %d\n", ccode);
	retcode = 3;
	goto Exit;

StatError:
	printf ("ERROR:  stat() returned error on temporary file %s\n", TEMP_FILE_NAME);
	retcode = 4;
	goto Exit;

OpenError:
	printf ("ERROR:  Cannot open temporary file %s for reading\n", TEMP_FILE_NAME);
	retcode = 5;
	goto Exit;

ReadError:
	printf ("ERROR:  Cannot read temporary file %s\n", TEMP_FILE_NAME);
	retcode = 6;
	goto Exit;
}	

int GetMore (int GetMoreOffset, char *OutputBuffer, int OutputBufferMaxSize, int *OutputBufferSize)
{
	int retcode;
	int ccode;
	int fd;
	struct stat filestat;

	// Assume failure
	retcode = -1;
	fd = -1;
	*OutputBufferSize = 0;

	// Determine the length of the temp file
	ccode = stat (TEMP_FILE_NAME, &filestat);
	if (ccode == -1)
		goto StatError;

	// Read the results file
	fd = open (TEMP_FILE_NAME, O_RDONLY);
	if (fd == -1)
		goto OpenError;
	ccode = lseek (fd, GetMoreOffset, SEEK_SET);
	if (ccode == -1)
		goto SeekError;
	ccode = read (fd, OutputBuffer, OutputBufferMaxSize);
	if (ccode == -1)
		goto ReadError;
	*OutputBufferSize = ccode;

	// OK
	if (filestat.st_size > OutputBufferMaxSize+GetMoreOffset)
		retcode = 1;	// Indicate that client needs to do GET_MORE commands
	else
		retcode = 0;

Exit:
	// Deallocate resources and return
	if (fd != -1)
		close (fd);
	return (retcode);

StatError:
	printf ("ERROR:  stat() returned error on temporary file %s\n", TEMP_FILE_NAME);
	retcode = 2;
	goto Exit;

OpenError:
	printf ("ERROR:  Cannot open temporary file %s for reading\n", TEMP_FILE_NAME);
	retcode = 3;
	goto Exit;

SeekError:
	printf ("ERROR:  Cannot seek in temporary file %s\n", TEMP_FILE_NAME);
	retcode = 4;
	goto Exit;

ReadError:
	printf ("ERROR:  Cannot read temporary file %s\n", TEMP_FILE_NAME);
	retcode = 5;
	goto Exit;
}	